Index: ip_input.c
===================================================================
RCS file: /home/ncvs/src/sys/netinet/ip_input.c,v
retrieving revision 1.270
diff -u -p -r1.270 ip_input.c
--- ip_input.c	2 May 2004 15:10:16 -0000	1.270
+++ ip_input.c	4 May 2004 09:11:37 -0000
@@ -1713,7 +1725,7 @@ ip_forward(struct mbuf *m, int srcrt, st
 	struct ip *ip = mtod(m, struct ip *);
 	struct in_ifaddr *ia;
 	int error, type = 0, code = 0;
-	struct mbuf *mcopy;
+	struct mbuf *mcopy = NULL;
 	n_long dest;
 	struct in_addr pkt_dst;
 	struct ifnet *destifp;
@@ -1757,6 +1769,7 @@ ip_forward(struct mbuf *m, int srcrt, st
 		return;
 	}
 
+#if 0 	/* mcopy hack */
 	/*
 	 * Save the IP header and at most 8 bytes of the payload,
 	 * in case we need to generate an ICMP message to the src.
@@ -1789,6 +1802,7 @@ ip_forward(struct mbuf *m, int srcrt, st
 		    (int)ip->ip_len);
 		m_copydata(m, 0, mcopy->m_len, mtod(mcopy, caddr_t));
 	}
+#endif	/* mcopy hack */
 
 #ifdef IPSTEALTH
 	if (!ipstealth) {
@@ -1856,7 +1870,14 @@ ip_forward(struct mbuf *m, int srcrt, st
 		*(struct sockaddr_in **)(mtag+1) = next_hop;
 		m_tag_prepend(m, mtag);
 	}
-	error = ip_output(m, (struct mbuf *)0, NULL, IP_FORWARDING, 0, NULL);
+	error = ip_output(m, (struct mbuf *)0, NULL,
+			  (IP_FORWARDING | IP_KEEPMBUF), 0, NULL);
+
+	if (error < 0) {
+		error -= error * 2;
+		mcopy = m;
+	}
+
 	if (error)
 		ipstat.ips_cantforward++;
 	else {
Index: ip_output.c
===================================================================
RCS file: /home/ncvs/src/sys/netinet/ip_output.c,v
retrieving revision 1.217
diff -u -p -r1.217 ip_output.c
--- ip_output.c	2 May 2004 15:10:16 -0000	1.217
+++ ip_output.c	4 May 2004 09:11:37 -0000
@@ -1065,7 +1078,18 @@ done:
 #endif
 	return (error);
 bad:
-	m_freem(m);
+	/*
+	 * When the flag IP_KEEPMBUF is set we try not to free the
+	 * mbuf if possible.  However there are some cases where an
+	 * mbuf is free'd anyway and this is a way to disinguish them.
+	 * A negative errno tells the caller that the mbuf is intact
+	 * and it can use it for icmp_error purposes.  The caller
+	 * should revert this for further processing of the errno.
+	 */
+	if (flags & IP_KEEPMBUF)
+		error -= error * 2;
+	else
+		m_freem(m);
 	goto done;
 }
 
@@ -1077,6 +1101,10 @@ bad:
  *
  * if_hwassist_flags is the hw offload capabilities (see if_data.ifi_hwassist)
  * sw_csum contains the delayed checksums flags (e.g., CSUM_DELAY_IP).
+ *
+ * Note: ip_fragment expects ip_len and ip_off in the supplied mbuf to be in
+ * host byte order.  However the returned fragment chain will be entirely in
+ * network byte order.
  */
 int
 ip_fragment(struct ip *ip, struct mbuf **m_frag, int mtu,
Index: ip_var.h
===================================================================
RCS file: /home/ncvs/src/sys/netinet/ip_var.h,v
retrieving revision 1.86
diff -u -p -r1.86 ip_var.h
--- ip_var.h	2 May 2004 06:36:30 -0000	1.86
+++ ip_var.h	4 May 2004 09:11:37 -0000
@@ -130,8 +130,9 @@ struct	ipstat {
 #define	IP_FORWARDING		0x1		/* most of ip header exists */
 #define	IP_RAWOUTPUT		0x2		/* raw ip header exists */
 #define	IP_SENDONES		0x4		/* send all-ones broadcast */
-#define	IP_ROUTETOIF		SO_DONTROUTE	/* bypass routing tables */
-#define	IP_ALLOWBROADCAST	SO_BROADCAST	/* can send broadcast packets */
+#define	IP_KEEPMBUF		0x8		/* do not free MBUF on error */
+#define	IP_ROUTETOIF		SO_DONTROUTE	/* (0x10) bypass routing tables */
+#define	IP_ALLOWBROADCAST	SO_BROADCAST	/* (0x20) can send broadcast packets */
 
 /* mbuf flag used by ip_fastfwd */
 #define	M_FASTFWD_OURS		M_PROTO1	/* changed dst to local */
